package com.m20.absolute.schematron;

import java.math.BigDecimal;
import java.util.ArrayList;
import java.util.List;
import java.util.Map;

import net.sourceforge.nrl.parser.ast.IModelReference;
import net.sourceforge.nrl.parser.ast.constraints.IArithmeticExpression;
import net.sourceforge.nrl.parser.ast.constraints.IBinaryOperatorStatement;
import net.sourceforge.nrl.parser.ast.constraints.IBinaryPredicate;
import net.sourceforge.nrl.parser.ast.constraints.IBooleanLiteral;
import net.sourceforge.nrl.parser.ast.constraints.IConstraint;
import net.sourceforge.nrl.parser.ast.constraints.IConstraintRuleDeclaration;
import net.sourceforge.nrl.parser.ast.constraints.IDecimalNumber;
import net.sourceforge.nrl.parser.ast.constraints.IIntegerNumber;
import net.sourceforge.nrl.parser.ast.constraints.ILiteralString;
import net.sourceforge.nrl.parser.ast.constraints.impl.IfThenStatementImpl;
import net.sourceforge.nrl.parser.model.IAttribute;
import net.sourceforge.nrl.parser.model.IModelElement;

import com.m20.absolute.schematron.ast.Assertion;
import com.m20.absolute.schematron.ast.Rule;
import com.m20.absolute.schematron.ast.xpath.ArithExpr;
import com.m20.absolute.schematron.ast.xpath.BinaryOperator;
import com.m20.absolute.schematron.ast.xpath.BinaryPredicate;
import com.m20.absolute.schematron.ast.xpath.FunctionCall;
import com.m20.absolute.schematron.ast.xpath.IfStatem;
import com.m20.absolute.schematron.ast.xpath.Literal;
import com.m20.absolute.schematron.ast.xpath.Number;
import com.m20.absolute.schematron.ast.xpath.Path;
import com.m20.absolute.schematron.ast.xpath.PathExpr;
import com.m20.absolute.schematron.ast.xpath.Step;

/**
 * The class implements a generator that translates NRL constraints in XPath
 * expressions that throws an error if the NRL constraint is not handled and a
 * translator that generates Schematron rules from NRL rules.
 * 
 * @author Adriana Alexandru
 */

public class NRLSchematronGenerator {

	protected PathExpr translate(IConstraint constraint) {
		if (constraint instanceof IIntegerNumber) {
			IIntegerNumber nrlNumber = (IIntegerNumber) constraint;
			Number digit = new Number();
			digit.setValue(BigDecimal.valueOf(nrlNumber.getNumber()));
			return digit;
		}
		if (constraint instanceof ILiteralString) {
			ILiteralString nrlString = (ILiteralString) constraint;
			Literal literal = new Literal();
			literal.setValue(nrlString.getString());
			return literal;
		}
		if (constraint instanceof IDecimalNumber) {
			IDecimalNumber nrlDecimal = (IDecimalNumber) constraint;
			Number number = new Number();
			number.setValue(BigDecimal.valueOf(nrlDecimal.getNumber()));
			return number;
		}
		if (constraint instanceof IBooleanLiteral) {
			IBooleanLiteral nrlBoolean = (IBooleanLiteral) constraint;
			FunctionCall functionCall = new FunctionCall();
			if (nrlBoolean.getBoolean() == true) {
				functionCall.setName("true()");
				return functionCall;
			} else {
				functionCall.setName("false()");
				return functionCall;
			}
		}

		if (constraint instanceof IArithmeticExpression) {
			IArithmeticExpression nrlArith = (IArithmeticExpression) constraint;
			ArithExpr arithExpr = new ArithExpr();
			arithExpr.setLeft(translate(nrlArith.getLeft()));
			arithExpr.setRight(translate(nrlArith.getRight()));

			switch (nrlArith.getOperator()) {
			case PLUS:
				arithExpr.setOp(ArithExpr.Operator.PLUS);
				break;
			case MINUS:
				arithExpr.setOp(ArithExpr.Operator.MINUS);
				break;
			case MOD:
				arithExpr.setOp(ArithExpr.Operator.MOD);
				break;
			case DIV:
				arithExpr.setOp(ArithExpr.Operator.DIV);
				break;
			case TIMES:
				arithExpr.setOp(ArithExpr.Operator.TIMES);
				break;
			}
			return arithExpr;
		}
		if (constraint instanceof IBinaryPredicate) {
			IBinaryPredicate nrlPred = (IBinaryPredicate) constraint;
			BinaryPredicate binaryPred = new BinaryPredicate();
			binaryPred.setLeft(translate(nrlPred.getLeft()));
			binaryPred.setRight(translate(nrlPred.getRight()));

			switch (nrlPred.getPredicate()) {
			case EQUAL:
				binaryPred.setOp(BinaryPredicate.Operator.EQUAL);
				break;
			case GREATER:
				binaryPred.setOp(BinaryPredicate.Operator.GREATER);
				break;
			case GREATER_OR_EQUAL:
				binaryPred.setOp(BinaryPredicate.Operator.GREATER_OR_EQUAL);
				break;
			case LESS:
				binaryPred.setOp(BinaryPredicate.Operator.LESS);
				break;
			case LESS_OR_EQUAL:
				binaryPred.setOp(BinaryPredicate.Operator.LESS_OR_EQUAL);
				break;
			case NOT_EQUAL:
				binaryPred.setOp(BinaryPredicate.Operator.NOT_EQUAL);
				break;
			}
			return binaryPred;
		}
		if (constraint instanceof IBinaryOperatorStatement) {
			IBinaryOperatorStatement nrlBinary = (IBinaryOperatorStatement) constraint;
			BinaryOperator binaryOp = new BinaryOperator();
			binaryOp.setLeft(translate(nrlBinary.getLeft()));
			binaryOp.setRight(translate(nrlBinary.getRight()));

			switch (nrlBinary.getOperator()) {
			case AND:
				binaryOp.setOp(BinaryOperator.Operator.AND);
				break;
			case OR:
				binaryOp.setOp(BinaryOperator.Operator.OR);
				break;
			case IFF:
				binaryOp.setOp(BinaryOperator.Operator.IFF);
				break;
			case IMPLIES:
				FunctionCall functionNot = new FunctionCall();
				ArrayList<PathExpr> param = new ArrayList<PathExpr>();
				param.add(binaryOp.getLeft());
				functionNot.addParamsToFunct("not", param);
				binaryOp.setLeft(functionNot);
				binaryOp.setOp(BinaryOperator.Operator.OR);
				break;
			}
			return binaryOp;
		}
		if (constraint instanceof IfThenStatementImpl) {
			IfThenStatementImpl nrlIfThen = (IfThenStatementImpl) constraint;
			IfStatem ifThenStatem = new IfStatem();
			ifThenStatem.setIfPart(translate(nrlIfThen.getIf()));
			ifThenStatem.setThenPart(translate(nrlIfThen.getThen()));
			return ifThenStatem;
		}

		if (constraint instanceof IModelReference) {
			IModelReference nrlModel = (IModelReference) constraint;
			Path path = new Path();
			Step step = new Step();
			step.setStepName(((IAttribute) nrlModel.getInitialStep()).getName());
			path.addStep(step);

			// if(!nrlModel.getRemainingSteps().isEmpty()){
			for (IAttribute rs : nrlModel.getRemainingSteps()) {
				Step s = new Step();
				s.setStepName(((IAttribute) rs).getName());
				path.addStep(s);
			}
			// }
			return path;
		}

		else
			throw new RuntimeException("This constraint " + constraint
					+ " is not handled.");
	}

	/**
	 * Generates a Schematron Rule from an NRL Rule
	 * 
	 * @param nrlRule
	 *            is an NRL rule
	 * @param pathMap
	 *            is a map of NRL Type to all possible XML paths
	 * @return a Schematron {@link Rule}
	 */
	public Rule translate(IConstraintRuleDeclaration nrlRule,
			Map<IModelElement, List<String>> pathMap) {
		Rule rule = new Rule();
		rule.setContext(pathMap.get(nrlRule.getContext()));
		rule.setId(nrlRule.getId());
		Assertion assertion = new Assertion();
		assertion.setTest(translate(nrlRule.getConstraint()));
		rule.setAssertion(assertion);

		return rule;
	}
}
